var protocol = require('./protocol')
var bitfield = require('sparse-bitfield')
var set = require('unordered-set')
var rle = require('bitfield-rle')

module.exports = replicate

function replicate (feed, opts) {
  var peer = new Peer(feed, opts)
  var stream = protocol({state: peer})

  peer.feed = feed
  peer.stream = stream

  stream.on('close', function () {
    peer.destroy()
  })

  feed.ready(function (err) {
    if (err) return stream.destroy(err)

    stream.setTimeout(5000)
    stream.id = feed.id

    stream.onhandshake = function (handshake) {
      stream.remoteId = handshake.id
      stream.emit('handshake')
    }

    stream.onwant = function (want, peer) {
      peer.onwant(want)
    }

    stream.onhave = function (have, peer) {
      peer.onhave(have)
    }

    stream.onrequest = function (request, peer) {
      peer.onrequest(request)
    }

    stream.ondata = function (data, peer) {
      peer.ondata(data)
    }

    stream.ontick = function (peer) {
      peer.ontick()
    }

    stream.onclose = function (close, peer) {
      peer.onclose()
    }

    stream.ontimeout = function () {
      peer.destroy(new Error('Stream timeout'))
    }

    peer.ready()
  })

  return stream
}

function Peer (feed, opts) {
  this.feed = feed
  this.stream = null // set by replicate just after creation
  this.remoteBitfield = null
  this.remoteLength = 0
  this.live = !!opts.live

  this.downloading = !feed.writable
  this.uploading = true

  this.maxRequests = opts.maxRequests || feed.maxRequests || 16
  this.inflightRequests = []

  this._index = -1
  this._lastBytes = 0
  this._closed = false
}

Peer.prototype.onwant = function () {
  // TODO: reply to the actual want context
  var rle = this.feed.bitfield.compress()
  this.stream.have({start: 0, bitfield: rle})
}

Peer.prototype.ondata = function (data) {
  var self = this

  this.feed._putBuffer(data.index, data.value, data, this, function (err) {
    if (err) return self.destroy(err)
    self._clear(data.index, !data.value)
  })
}

Peer.prototype._clear = function (index, hash) {
  // TODO: optimize me (no splice and do not run through all ...)
  for (var i = 0; i < this.inflightRequests.length; i++) {
    if (this.inflightRequests[i].index === index) {
      this.inflightRequests.splice(i, 1)
      i--
    }
  }

  if (hash) {
    // TODO: once we compress the _reserved bitfield we should clear it always here ...
    // since we don't right now nothing is gained unless it is a hash response
    this.feed._reserved.set(index, false)
  }

  // TODO: only update all if we have overlapping selections
  this.feed._updatePeers()
}

Peer.prototype.onrequest = function (request) {
  if (request.bytes) return this._onbytes(request)

  var self = this
  var opts = {digest: request.nodes, hash: request.hash}

  this.feed.proof(request.index, opts, onproof)

  function onproof (err, proof) {
    if (err) return self.destroy(err)
    if (request.hash) onvalue(null, null)
    else self.feed._getBuffer(request.index, onvalue)

    function onvalue (err, value) {
      if (err) return self.destroy(err)

      if (value) {
        if (!self.remoteBitfield.set(request.index, true)) return
        self.feed.emit('upload', request.index, value, self)
      } else {
        if (self.remoteBitfield.get(request.index)) return
      }

      self.stream.data({
        index: request.index,
        value: value,
        nodes: proof.nodes,
        signature: proof.signature
      })
    }
  }
}

Peer.prototype._onbytes = function (request) {
  var self = this

  this.feed.seek(request.bytes, {wait: false}, function (err, index) {
    if (err) {
      request.bytes = 0
      self.onrequest(request)
      return
    }

    // quick'n'dirty filter for parallel bytes requests
    // it does not matter that this doesn't catch ALL parallel requests - just a bandwidth optimization
    if (self._lastBytes === request.bytes) return
    self._lastBytes = request.bytes

    request.bytes = 0
    request.index = index
    request.nodes = 0

    self.onrequest(request)
  })
}

Peer.prototype.ontick = function () {
  if (!this.inflightRequests.length) return

  var first = this.inflightRequests[0]
  if (--first.tick) return

  if (first.hash ? this.feed.tree.get(2 * first.index) : this.feed.bitfield.get(first.index)) {
    // prob a bytes response
    this.inflightRequests.shift()
    this.feed._reserved.set(first.index, false)
    return
  }

  this.destroy(new Error('Request timeout'))
}

Peer.prototype.onhave = function (have) {
  if (have.bitfield) { // TODO: handle start !== 0
    this.remoteBitfield = bitfield(rle.decode(have.bitfield))
    if (this.remoteBitfield.length > this.remoteLength) {
      this.remoteLength = this.remoteBitfield.length
      while (this.remoteLength && !this.remoteBitfield.get(this.remoteLength - 1)) this.remoteLength--
    }
  } else {
    if (!this.remoteBitfield) this.remoteBitfield = bitfield()
    // TODO: if len > something simply copy a 0b1111... buffer to the bitfield

    var start = have.start
    var len = have.length || 1

    while (len--) this.remoteBitfield.set(start++, true)
    if (start > this.remoteLength) this.remoteLength = start
  }

  this._updateEnd()
  this.update()
}

Peer.prototype._updateEnd = function () {
  if (this.live || this.feed.sparse || !this.feed._selections.length) return

  var sel = this.feed._selections[0]
  var remoteLength = -1

  for (var i = 0; i < this.feed.peers.length; i++) {
    if (this.feed.peers[i].remoteLength > remoteLength) {
      remoteLength = this.feed.peers[i].remoteLength
    }
  }

  sel.end = remoteLength
}

Peer.prototype.onclose = function () {
  this.update()
  if (this.feed._selections.length && this.downloading) return
  this.end()
}

Peer.prototype.have = function (have) { // called by feed
  this.stream.have(have)
}

Peer.prototype.haveBytes = function (bytes) { // called by feed
  for (var i = 0; i < this.inflightRequests.length; i++) {
    if (this.inflightRequests[i].bytes === bytes) {
      this.feed._reserved.set(this.inflightRequests[i].index, false)
      this.inflightRequests.splice(i, 1)
      i--
    }
  }

  this.update()
}

Peer.prototype.update = function () {
  if (!this.downloading || !this.remoteBitfield) return

  var selections = this.feed._selections
  var waiting = this.feed._waiting
  var inflight = this.inflightRequests.length
  var offset = 0
  var i = 0

  // TODO: less duplicate code here
  // TODO: re-add priority levels

  while (inflight < this.maxRequests) {
    offset = Math.floor(Math.random() * waiting.length)

    for (i = 0; i < waiting.length; i++) {
      var w = waiting[offset++]
      if (offset === waiting.length) offset = 0

      this._downloadWaiting(w)
    }
    if (inflight === this.inflightRequests.length) break
    inflight = this.inflightRequests.length
  }

  while (inflight < this.maxRequests) {
    offset = Math.floor(Math.random() * selections.length)

    for (i = 0; i < selections.length; i++) {
      var s = selections[offset++]
      if (offset === selections.length) offset = 0

      if (!s.iterator) s.iterator = this.feed.bitfield.iterator(s.start, s.end)
      this._downloadRange(s)
    }

    if (inflight === this.inflightRequests.length) return
    inflight = this.inflightRequests.length
  }
}

Peer.prototype.ready = function () {
  set.add(this.feed.peers, this)
  this.stream.ready(this.feed.key)
  this.stream.handshake({id: this.feed.id})
  this.stream.want({start: 0}) // TODO: don't just subscribe to *EVERYTHING* hehe
}

Peer.prototype.end = function () {
  if (!this._closed) {
    this._closed = true
    this.stream.close()
  } else {
    this.stream.push(null)
  }
}

Peer.prototype.destroy = function (err) {
  if (this._index === -1) return
  set.remove(this.feed.peers, this)
  for (var i = 0; i < this.inflightRequests.length; i++) {
    this.feed._reserved.set(this.inflightRequests[i].index, false)
  }
  this._updateEnd()
  this._index = -1
  this.stream.destroy(err)
  this.feed._updatePeers()
}

Peer.prototype._downloadWaiting = function (wait) {
  if (!wait.bytes) {
    if (!this.remoteBitfield.get(wait.index) || !this.feed._reserved.set(wait.index, true)) return
    this._request(wait.index, 0, false)
    return
  }

  this._downloadRange(wait)
}

Peer.prototype._downloadRange = function (range) {
  if (!range.iterator) range.iterator = this.feed.bitfield.iterator(range.start, range.end)

  var reserved = this.feed._reserved
  var ite = range.iterator
  var wantedEnd = Math.min(range.end === -1 ? this.remoteLength : range.end, this.remoteLength)

  if (ite.end !== wantedEnd) ite.range(range.start, wantedEnd)

  var i = range.linear ? ite.next() : ite.random()
  var reset = false
  var start = i

  if (i === -1) {
    if (!range.bytes && ite.seek(range.start).next() === -1 && (range.end > -1 && this.remoteLength >= range.end)) {
      set.remove(this.feed._selections, range)
      range.callback(null)
      if (!this.live && !this.sparse && !this.feed._selections.length) this.end()
    }
    return
  }

  while (!this.remoteBitfield.get(i) || (range.hash && this.feed.tree.get(2 * i)) || !reserved.set(i, true)) {
    i = ite.next()
    reset = true

    if (i > -1) {
      // check this index
      continue
    }

    if (!range.linear && start !== 0) {
      // retry from the beginning since we are iterating randomly and started !== 0
      i = ite.seek(0).next()
      start = 0
      continue
    }

    // we have checked all indexes.
    // if we are looking for hashes we should check if we have all now (first check only checks blocks)
    if (range.hash) {
      // quick'n'dirty check if have all hashes - can be optimized be checking only tree roots
      // but we don't really request long ranges of hashes so yolo
      for (var j = range.start; j < wantedEnd; j++) {
        if (!this.feed.tree.get(2 * j)) return
      }
      if (!range.bytes) {
        set.remove(this.feed._selections, range)
        range.callback(null)
      }
    }

    // exit the update loop - nothing to do
    return
  }

  if (reset) ite.seek(range.start)

  this._request(i, range.bytes || 0, range.hash)
}

Peer.prototype._request = function (index, bytes, hash) {
  var request = {
    tick: 6,
    bytes: bytes,
    index: index,
    hash: hash,
    nodes: this.feed.digest(index)
  }

  this.inflightRequests.push(request)
  this.stream.request(request)
}
